/*
 * This file is part of RTRlib.
 *
 * This file is subject to the terms and conditions of the MIT license.
 * See the file LICENSE in the top level directory for more details.
 *
 * Website: http://rtrlib.realmv6.org/
 */

#ifndef SSH_TRANSPORT_PRIVATE_H
#define SSH_TRANSPORT_PRIVATE_H
#include "rtrlib/transport/ssh/ssh_transport.h"
#endif
